%{
/*
 * The contents of this file are subject to the Spread Non-Commercial 
 * License, Version 1.2 (the ``License''); you may not use
 * this file except in compliance with the License.  You may obtain a
 * copy of the License at:
 *
 * http://www.spread.org/license/
 *
 * or in the file ``license.txt'' found in this distribution.
 *
 * Software distributed under the License is distributed on
 * an AS IS basis, WITHOUT WARRANTY OF ANY KIND,
 * either express or implied. See the License for the
 * specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is:
 *    The Spread Wide Area Group Communication Toolkit.
 *     
 * The Initial Developers of the Original Code are:
 *    Yair Amir, Michal Miskin-Amir, Jonathan Stanton.
 *
 *    Portions created by the Initial Developers are
 *    Copyright (C) 1993-2000 Yair Amir <yairamir@cs.jhu.edu>,
 *     			      Michal Miskin-Amir <michal@spread.org>,
 *			      Jonathan Stanton <jonathan@cs.jhu.edu>.
 *    All Rights Reserved.
 *
 * Major Contributor(s):
 * ---------------
 *    Dan Schoenblum   dansch@cnds.jhu.edu - Java Interface Developer.
 *    John Schultz     jschultz@cnds.jhu.edu - contribution to process group membership.
 *    Theo Schlossnagle theos@cnds.jhu.edu - Perl library and Skiplists.
 *
 */

#include "arch.h"
#include <stdio.h>
#include <string.h>

#ifndef ARCH_PC_WIN95
#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/param.h>

#else /* ARCH_PC_WIN95 */
#include <winsock.h>
#endif /* ARCH_PC_WIN95 */

#include "alarm.h"
#include "configuration.h"
#include "skiplist.h"
#include "memory.h"
#include "objects.h"
#include "conf_body.h"

        int     line_num, semantic_errors;
 extern char    *yytext;
 extern int     yyerror(char *str);
 extern void    yywarn(char *str);
 extern int     yylex();

 static int     num_procs = 0;
 static int     segment_procs = 0;
 static int     segments = 0;
 static int     rvec_num = 0;
 static proc    *new_proc;
 static int32   temp_seg_list[MAX_PROCS_SEGMENT];

static char *segment2str(int seg) {
  static char ipstr[40];
  int id = Config.segments[seg].bcast_address;
  sprintf(ipstr, "%d.%d.%d.%d:%d",
  	(id & 0xff000000)>>24,
  	(id & 0xff0000)>>16,
  	(id & 0xff00)>>8,
  	(id & 0xff),
	Config.segments[seg].port);
  return ipstr;
}
static void alarm_print_proc(proc *p, int port) {
  if(port == p->port)
    Alarm(CONF, "\t%20s: %d.%d.%d.%d\n", p->name,
  	  (p->id & 0xff000000)>>24,
  	  (p->id & 0xff0000)>>16,
  	  (p->id & 0xff00)>>8,
  	  (p->id & 0xff));
  else
    Alarm(CONF, "\t%20s: %d.%d.%d.%d:%d\n", p->name,
  	  (p->id & 0xff000000)>>24,
  	  (p->id & 0xff0000)>>16,
  	  (p->id & 0xff00)>>8,
  	  (p->id & 0xff),
	  p->port);
}

static int32u name2ip(char *name) {
  int anip, i1, i2, i3, i4;
  struct hostent *host_ptr;

  host_ptr = gethostbyname(name);
  
  if ( host_ptr == 0)
    Alarm( EXIT, "Conf_init: no such host %s\n",
	   name);
  
  memcpy(&anip, host_ptr->h_addr_list[0], 
	 sizeof(int32) );
  anip = htonl( anip );
  i1= ( anip & 0xff000000 ) >> 24;
  i2= ( anip & 0x00ff0000 ) >> 16;
  i3= ( anip & 0x0000ff00 ) >>  8;
  i4=   anip & 0x000000ff;
  return ((i1<<24)|(i2<<16)|(i3<<8)|i4);
}

static  void expand_filename(char *out_string, int str_size, const char *in_string)
{
  const char *in_loc;
  char *out_loc;
  char hostn[256];
  
  for ( in_loc = in_string, out_loc = out_string; out_loc - out_string < str_size; in_loc++ )
  {
          if (*in_loc == '%' ) {
                  switch( in_loc[1] ) {
                  case 'h':
                  case 'H':
                          gethostname(hostn, sizeof(hostn) );
                          out_loc += snprintf(out_loc, str_size - (out_loc - out_string), "%s", hostn); 
                          in_loc++;
                          continue;
                  default:
                          break;
                  }

          }
          *out_loc = *in_loc;
          out_loc++;
          if (*in_loc == '\0') break;
  }
  out_string[str_size] = '\0';
}

#define PROCS_CHECK( num_procs, stoken ) { \
                                            char strbuf[80]; \
                                            if ( (num_procs) >= MAX_PROCS_RING ) { \
                                                snprintf(strbuf, 80, "%s (Too many daemons configured--%d max)\n", (stoken), MAX_PROCS_RING); \
                                                return (yyerror(strbuf)); \
                                            } \
                                         }
#define SEGMENT_CHECK( num_segments, stoken )  { \
                                            char strbuf[80]; \
                                            if ( (num_segments) >= MAX_SEGMENTS ) { \
                                                snprintf(strbuf, 80, "%s (Too many segments configured--%d max)\n", (stoken), MAX_SEGMENTS); \
                                                return( yyerror(strbuf)); \
                                            } \
                                         }
#define SEGMENT_SIZE_CHECK( num_procs, stoken )  { \
                                            char strbuf[80]; \
                                            if ( (num_procs) >= MAX_PROCS_SEGMENT ) { \
                                                snprintf(strbuf, 80, "%s (Too many daemons configured in segment--%d max)\n", (stoken), MAX_PROCS_SEGMENT); \
                                                return( yyerror(strbuf)); \
                                            } \
                                         }


%}
%start Config
%token SEGMENT EVENTLOGFILE EVENTTIMESTAMP IPADDR NUMBER COLON
%token OPENBRACE CLOSEBRACE EQUALS STRING
%token DEBUGFLAGS BANG
%token DDEBUG DEXIT DPRINT DDATA_LINK DNETWORK DPROTOCOL DSESSION
%token DCONF DMEMB DFLOW_CONTROL DSTATUS DEVENTS DGROUPS DMEMORY
%token DSKIPLIST DALL DNONE
%token DANGEROUSMONITOR SP_BOOL LINKPROTOCOL PHOP PTCPHOP
%token ROUTEMATRIX LINKCOST
%%
Config		:	ConfigStructs
                        {
			  Config.num_segments = segments;
			  Num_procs = num_procs;
			  Alarm(CONF, "Finished configuration file.\n");
			}


ConfigStructs	:	SegmentStruct ConfigStructs
		|	ParamStruct ConfigStructs
		|	RouteStruct ConfigStructs
		|
		;

AlarmBit	:	DDEBUG { $$ = $1; }
		|	DEXIT { $$ = $1; }
		|	DPRINT { $$ = $1; }
		|	DDATA_LINK { $$ = $1; }
		|	DNETWORK { $$ = $1; }
		|	DPROTOCOL { $$ = $1; }
		|	DSESSION { $$ = $1; }
		|	DCONF { $$ = $1; }
		|	DMEMB { $$ = $1; }
		|	DFLOW_CONTROL { $$ = $1; }
		|	DSTATUS { $$ = $1; }
		|	DEVENTS { $$ = $1; }
		|	DGROUPS { $$ = $1; }
		|	DMEMORY { $$ = $1; }
		|	DSKIPLIST { $$ = $1; }
		|	DALL { $$ = $1; }
		|	DNONE { $$ = $1; }
		;

AlarmBitComp	:	AlarmBitComp AlarmBit
			{
			  $$.mask = ($1.mask | $2.mask);
			}
		|	AlarmBitComp BANG AlarmBit
			{
			  $$.mask = ($1.mask & ~($3.mask));
			}
		|	{ $$.mask = NONE; }
		;

ParamStruct	:	DEBUGFLAGS EQUALS OPENBRACE AlarmBitComp CLOSEBRACE
			{
			  if (! Alarm_get_interactive() ) {
                            Alarm_clear(ALL);
			    Alarm_set($4.mask);
			    Alarm(CONF, "Set Alarm mask to: %x\n", Alarm_get());
                          }
			}
		|	EVENTLOGFILE EQUALS STRING
			{
			  if (! Alarm_get_interactive() ) {
                            char file_buf[MAXPATHLEN];
                            expand_filename(file_buf, MAXPATHLEN, $3.string);
                            Alarm_set_output(file_buf);
                          }
			}
		|	EVENTTIMESTAMP EQUALS STRING
			{
			  if (! Alarm_get_interactive() ) {
			    Alarm_enable_timestamp($3.string);
                          }
			}
		|	EVENTTIMESTAMP
			{
			  if (! Alarm_get_interactive() ) {
			    Alarm_enable_timestamp(NULL);
                          }
			}
		|	DANGEROUSMONITOR EQUALS SP_BOOL
			{
			  if (! Alarm_get_interactive() ) {
                            Conf_set_dangerous_monitor_state($3.boolean);
                          }
			}
		|	LINKPROTOCOL EQUALS PHOP
			{
                            Conf_set_link_protocol(HOP_PROT);
			}
		|	LINKPROTOCOL EQUALS PTCPHOP
			{
                            Conf_set_link_protocol(TCP_PROT);
			}

SegmentStruct	:    SEGMENT IPADDR OPENBRACE Segmentparams CLOSEBRACE
                        { int i;
                          SEGMENT_CHECK( segments, inet_ntoa($2.ip.addr) );
			  Config.segments[segments].num_procs = segment_procs;
			  Config.segments[segments].port = $2.ip.port;
			  Config.segments[segments].bcast_address =
			    $2.ip.addr.s_addr;
			  if(Config.segments[segments].port == 0)
			    Config.segments[segments].port = DEFAULT_SPREAD_PORT;
			  Alarm(CONF, "Successfully configured Segment %d [%s] with %d procs:\n",
				segments,
				segment2str(segments),
				segment_procs);
			  for(i=(num_procs - segment_procs); i<num_procs; i++) {
                                  /* This '1' is to keep each proc with the same port as the segment.*/
			    if( 1 || Config_procs[i].port==0)  {
			      Config_procs[i].port=
				Config.segments[segments].port;
			    }
			    alarm_print_proc(&Config_procs[i],
			    	Config.segments[segments].port);
			  }
			  segments++;
			  segment_procs = 0;
			}
		;

Segmentparams	:	Segmentparam Segmentparams
		|
		;

Segmentparam	:	STRING IPADDR
			{ 
                          PROCS_CHECK( num_procs, $1.string );
                          SEGMENT_CHECK( segments, $1.string );
                          SEGMENT_SIZE_CHECK( segment_procs, $1.string );
                          strcpy(Config_procs[num_procs].name, $1.string);
                          Config_procs[num_procs].id = $2.ip.addr.s_addr;
 		          Config_procs[num_procs].port = $2.ip.port;
			  Config_procs[num_procs].seg_index = segments;
			  Config_procs[num_procs].index_in_seg = segment_procs;
			  Config.segments[segments].proc_ids[segment_procs] = 
			    Config_procs[num_procs].id;
			  num_procs++;
			  segment_procs++;
			}
		|	STRING
			{ 
                          PROCS_CHECK( num_procs, $1.string );
                          SEGMENT_CHECK( segments, $1.string );
                          SEGMENT_SIZE_CHECK( segment_procs, $1.string );
                          strcpy(Config_procs[num_procs].name, $1.string);
                          Config_procs[num_procs].id =
			    name2ip(Config_procs[num_procs].name);
 		          Config_procs[num_procs].port = 0;
			  Config_procs[num_procs].seg_index = segments;
			  Config_procs[num_procs].index_in_seg = segment_procs;
			  Config.segments[segments].proc_ids[segment_procs] = 
			    Config_procs[num_procs].id;
			  num_procs++;
			  segment_procs++;
			}
		;

RouteStruct	:    ROUTEMATRIX OPENBRACE Routevectors CLOSEBRACE
                        { 
			  Alarm(CONF, "Successfully configured Routing Matrix for %d Segments with %d rows in the routing matrix\n",segments, rvec_num);
			}
		;

Routevectors	:	Routevectors Routevector
		|
		;

Routevector     :       LINKCOST
                        { 
                                int rvec_element;
                                for (rvec_element = 0; rvec_element < segments; rvec_element++) {
                                        if ($1.cost[rvec_element] < 0) yyerror("Wrong number of entries for routing matrix");
                                        LinkWeights[rvec_num][rvec_element] = $1.cost[rvec_element];
                                }
                                rvec_num++;
                        }
                |       
                ;
%%
void yywarn(char *str) {
        fprintf(stderr, "-------Parse Warning-----------\n");
        fprintf(stderr, "Parser warning on or before line %d\n", line_num);
        fprintf(stderr, "Error type; %s\n", str);
        fprintf(stderr, "Offending token: %s\n", yytext);
}
int yyerror(char *str) {
  fprintf(stderr, "-------------------------------------------\n");
  fprintf(stderr, "Parser error on or before line %d\n", line_num);
  fprintf(stderr, "Error type; %s\n", str);
  fprintf(stderr, "Offending token: %s\n", yytext);
  exit(1);
}
