

package spread;

/**
 * Objects of a class that implements the AdvancedMessageListener interface can
 * add themselves to a SpreadConnection with {@link SpreadConnection#add(AdvancedMessageListener)}.
 * The object is an active listener until it is removed by a call to 
 * {@link SpreadConnection#remove(AdvancedMessageListener)}.  While the listener is active, it will be alerted
 * to all messages received on the connection.  When a regular message is received, 
 * {@link AdvancedMessageListener#regularMessageReceived(SpreadMessage)} will be called.
 * When a membership message is received, {@link AdvancedMessageListener#membershipMessageReceived(SpreadMessage)}
 * will be called.
 */
public interface AdvancedMessageListener
{
	// A new regular message has been receieved.
	////////////////////////////////////////////
	/**
	 * If the object has been added to a connection with {@link SpreadConnection#add(AdvancedMessageListener)},
	 * this gets called whenever a regular message is received.  The call happens in a seperate thread.
	 * 
	 * @param  message  the message that has been received
	 */
	public void regularMessageReceived(SpreadMessage message);
	
	// A new membership message has been received.
	//////////////////////////////////////////////
	/**
	 * If the object has been added to a connection with {@link SpreadConnection#add(AdvancedMessageListener)},
	 * this gets called whenever a membership message is received.  The call happens in a seperate thread.
	 * 
	 * @param  message  the message that has been received
	 */
	public void membershipMessageReceived(SpreadMessage message);
}
