

package spread;

/**
 * Objects of a class that implements the BasicMessageListener interface can
 * add themselves to a SpreadConnection with {@link SpreadConnection#add(BasicMessageListener)}.
 * The object is an active listener until it is removed by a call to 
 * {@link SpreadConnection#remove(BasicMessageListener)}.  While the listener is active, it will be alerted
 * to all messages received on the connection.  When a message is received, 
 * {@link BasicMessageListener#messageReceived(SpreadMessage)} will be called.
 */
public interface BasicMessageListener
{
	// A new message has been received.
	///////////////////////////////////
	/**
	 * If the object has been added to a connection with {@link SpreadConnection#add(BasicMessageListener)},
	 * this gets called whenever a message is received.  The call happens in a seperate thread.
	 * 
	 * @param  message  the message that has been received
	 */
	public void messageReceived(SpreadMessage message);
}
