

package spread;

/**
 * A SpreadException is thrown whenever a problem occurs in a spread method.
 * One example is calling a SpreadConnection object's {@link SpreadConnection#receive()} method
 * before calling {@link SpreadConnection#connect(InetAddress, int, String, boolean, boolean)}
 * on the object.
 * <p>
 * More exceptions will probably be provided in the future, to give more detailed error information.
 * However, these exceptions will all be extended from SpreadException, so any exception-handling
 * code written to handle SpreadException's will catch the new exceptions equally well.
 */
public class SpreadException extends Exception
{
	// Constructor.
	///////////////
	/**
	 * Creates a SpreadException with no error message.
	 */
	public SpreadException()
	{
		super();
	}
	
	// Constructor.
	///////////////
	/**
	 * Creates a SpreadException with an error message.
	 * 
	 * @param  errorMessage  a description of the error
	 */
	public SpreadException(String errorMessage)
	{
		super(errorMessage);
	}
}
