/*
 * The contents of this file are subject to the Spread Non-Commercial 
 * License, Version 1.3 (the ``License''); you may not use
 * this file except in compliance with the License.  You may obtain a
 * copy of the License at:
 *
 * http://www.spread.org/license/
 *
 * or in the file ``license.txt'' found in this distribution.
 *
 * Software distributed under the License is distributed on
 * an AS IS basis, WITHOUT WARRANTY OF ANY KIND,
 * either express or implied. See the License for the
 * specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is:
 *    The Spread Wide Area Group Communication Toolkit.
 *     
 * The Creators of the Original Code are:
 *    Yair Amir, Michal Miskin-Amir, Jonathan Stanton.
 *
 *    Portions created by the Initial Developers are
 *    Copyright (C) 1993-2000 Spread Concepts LLC <spread@spreadconcepts.com>
 *
 *    All Rights Reserved.
 *
 * Major Contributor(s):
 * ---------------
 *    Dan Schoenblum   dansch@cnds.jhu.edu - Java Interface Developer.
 *    John Schultz     jschultz@cnds.jhu.edu - contribution to process group membership.
 *    Theo Schlossnagle theos@cnds.jhu.edu - Perl library and Skiplists.
 *
 */


#include "arch.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#ifndef ARCH_PC_WIN95

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>

#ifdef  ARCH_SPARC_SOLARIS
#include <sys/uio.h>
#endif

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/un.h>
#include <signal.h>
#include <sys/ioctl.h>

#else   /* ARCH_PC_WIN95 */

#include <winsock.h>
#define	ioctl 	ioctlsocket

#endif  /* ARCH_PC_WIN95 */

#include "spread_params.h"
#include "net_types.h"
#include "sp_events.h"
#include "objects.h"
#include "memory.h"
#include "session.h"
#include "sess_types.h"

#define ext_sess_body 
#include "sess_body.h"
#undef  ext_sess_body

#include "groups.h"
#include "log.h"
#include "status.h"
#include "alarm.h"
#include "objects.h"
#include "prot_body.h"
#if     ( SPREAD_PROTOCOL > 3 )
#include "queues.h"
#endif
#include "message.h"

static	sp_time		Badger_timeout = { 0, 100000 };

static	message_obj	New_mess;

static	mailbox		Accept_inet_mbox;
static	mailbox		Accept_unix_mbox;

static	int		Protocol_threshold;

static	void	Sess_attach_accept(void);
static	void	Sess_detach_accept(void);
static	void    Sess_accept( mailbox mbox, int domain, void *dummy );
static	void	Sess_accept_continue( mailbox, int, void * );
static	void    Sess_read( mailbox mbox, int domain, void *dummy );
static	void	Sess_badger( mailbox mbox, void *dummy );
static	void	Sess_kill( mailbox mbox );
static	void	Sess_handle_join( message_link *mess_link );
static	void	Sess_handle_leave( message_link *mess_link );
static	void	Sess_handle_kill( message_link *mess_link );
static	void	Sess_attach_accept(void);
static	void	Sess_detach_accept(void);


void	Sess_init()
{
	struct	sockaddr_in	inet_addr;
	int16			port;
	int			ret;
	mailbox			mbox;

#ifndef ARCH_PC_WIN95

	struct	sockaddr_un	unix_addr;
	char			name[80];

	signal( SIGPIPE, SIG_IGN );

#endif	/* ARCH_PC_WIN95 */

        ret = Mem_init_object( MESSAGE_LINK, sizeof(message_link), 1000, 0);
        if (ret < 0)
        {
                Alarm(EXIT, "Sess_init: Failure to Initialize MESSAGE_LINK memory objects\n");
        }

        ret = Mem_init_object( DOWN_LINK, sizeof(down_link), 200, 0);
        if (ret < 0)
        {
                Alarm(EXIT, "Sess_Init: Failure to Initialize DOWN_LINK memory objects\n");
        }
        
	Num_sessions = 0;
	GlobalStatus.num_sessions = Num_sessions;
	GlobalStatus.message_delivered = 0;
	My 	 = Conf_my();
	port 	 = My.port;
	Session_threshold = LOW_PRIORITY;

	/* Initializing the protocol */
	Protocol_threshold = LOW_PRIORITY;

        Prot_init_down_queues();
        Prot_set_down_queue( NORMAL_DOWNQUEUE );
        Prot_init();

	/* Initiation of the INET socket */

	if( (mbox = socket( AF_INET, SOCK_STREAM, 0 ) ) == -1)
	    Alarm( EXIT, "Sess_init: INET sock error\n" );

	inet_addr.sin_family	= AF_INET;
	inet_addr.sin_addr.s_addr= INADDR_ANY;
	inet_addr.sin_port	= htons(port);

	if( bind( mbox,  (struct sockaddr *)&inet_addr, sizeof(inet_addr) ) == -1) 
	{
		Alarm( PRINT, "Sess_init: INET unable to bind to port %d, already running \n" ,port );
		exit(0);
	}
	Alarm( SESSION, "Sess_init: INET bind for port %d ok\n", port );

	if( listen( mbox, 5 ) < 0 ) 
	    Alarm( EXIT, "Sess_init: INET unable to listen\n" );

	Accept_inet_mbox = mbox;
        Alarm( SESSION, "Sess_init: INET went ok on mailbox %d\n", mbox );

#ifndef ARCH_PC_WIN95

	/* Initiation of the UNIX socket */

	if( (mbox = socket( AF_UNIX, SOCK_STREAM, 0 ) ) == -1)
	    Alarm( EXIT, "Sess_init: UNIX sock error\n" );

	unix_addr.sun_family	= AF_UNIX;
	sprintf( name, "/tmp/%d",My.port );
	strcpy( unix_addr.sun_path, name ); 
	unlink( name );

	if( bind( mbox, (struct sockaddr *)&unix_addr, sizeof(unix_addr) ) == -1) 
	{
		Alarm( PRINT, "Sess_init: UNIX unable to bind to name %s, already running \n" , name );
		exit(0);
	}
	Alarm( SESSION, "Sess_init: UNIX bind for name %s ok\n", name );

	chmod( name, 0666 );

	if( listen( mbox, 5 ) < 0 ) 
	    Alarm( EXIT, "Sess_init: UNIX unable to listen\n" );

	Accept_unix_mbox = mbox;
        Alarm( SESSION, "Sess_init: UNIX went ok on mailbox %d\n", mbox );

#endif	/* ARCH_PC_WIN95 */

	Sess_attach_accept();

        Message_populate_with_buffers(&New_mess);

	G_init();

        Alarm( SESSION, "Sess_init: ended ok\n" );
}

void	Sess_set_active_threshold()
{
	/* This function is used only by the session (and groups) layer */
	
	if( Protocol_threshold > Session_threshold ) 
		E_set_active_threshold( Protocol_threshold );
	else 	E_set_active_threshold( Session_threshold );
}

void    Sess_block_user(int xxx)
{

        Alarm(EXIT,"Sess_block_user: NOT IMPLEMENTED!\n");
}

void    Sess_unblock_user(int xxx)
{

        Alarm(EXIT,"Sess_unblock_user: NOT IMPLEMENTED!\n");
}
void	Sess_block_users_level()
{
	/* This function is used only by lower layers (protocol) */
	if( Protocol_threshold < MEDIUM_PRIORITY )
	{
		Protocol_threshold = MEDIUM_PRIORITY;
		Sess_set_active_threshold();
	}
}

void	Sess_unblock_users_level()
{
	/* This function is used only by lower layers (protocol) */
	if( Protocol_threshold > LOW_PRIORITY )
	{
		Protocol_threshold = LOW_PRIORITY;
		Sess_set_active_threshold();
	}
}

static	void	Sess_attach_accept()
{

	E_attach_fd( Accept_inet_mbox, READ_FD, Sess_accept, AF_INET, NULL, LOW_PRIORITY );
	E_attach_fd( Accept_inet_mbox, EXCEPT_FD, Sess_accept, AF_INET, NULL, LOW_PRIORITY );

#ifndef ARCH_PC_WIN95

	E_attach_fd( Accept_unix_mbox, READ_FD, Sess_accept, AF_UNIX, NULL, LOW_PRIORITY );

#endif	/* ARCH_PC_WIN95 */

}

static	void	Sess_detach_accept()
{

	E_detach_fd( Accept_inet_mbox, READ_FD );
	E_detach_fd( Accept_inet_mbox, EXCEPT_FD );

#ifndef ARCH_PC_WIN95

	E_detach_fd( Accept_unix_mbox, READ_FD );

#endif	/* ARCH_PC_WIN95 */

}

void	Sess_accept_continue2(int d1, void *d2)
{
	Sess_accept_continue(0,0,NULL);
} 
    

static	void	Sess_accept( mailbox mbox, int domain, void *dummy )
{
	struct	sockaddr_in	inet_addr;
	socklen_t		inet_len;
        sockopt_len_t           onlen;
	sp_time			accept_delay;
	char			response;
	int			ret;
	int			i;

	int32			on;

	if( domain == AF_INET ) 
	{
		inet_len = sizeof(inet_addr);
		Sessions[MAX_SESSIONS].mbox = accept( mbox, (struct sockaddr *)&inet_addr, &inet_len );

		Sessions[MAX_SESSIONS].type = AF_INET;
		/* 
		 * sender's machine ip address is: htonl(inet_addr.sin_addr.s_addr) 
		 * sender's assigned port is     : htons(inet_addr.sin_port)
		 */
		Sessions[MAX_SESSIONS].address = htonl(inet_addr.sin_addr.s_addr);

	}else if( domain == AF_UNIX ){
		/* no need for return values for AF_UNIX on the accept */
		Sessions[MAX_SESSIONS].mbox = accept( mbox, 0, 0 );

		Sessions[MAX_SESSIONS].type = AF_UNIX;
		Sessions[MAX_SESSIONS].address = 0;

	}else Alarm( EXIT, "Sess_accept: Unknown domain %d on mailbox %d\n", domain, mbox );


	if( Sessions[MAX_SESSIONS].mbox < 0 ) 
	{
		Alarm( SESSION, "Sess_accept: accept failed for domain %d\n", domain );
		return;
	}
	if( Num_sessions == MAX_SESSIONS )
	{
		response = REJECT_QUOTA;
		send( Sessions[MAX_SESSIONS].mbox, &response, sizeof(response), 0 );
		close( Sessions[MAX_SESSIONS].mbox );
		Alarm( SESSION, "Sess_accept: rejecting session due to quota\n" );
		return;
	}

	for( i=10; i <= 200; i+=5 )
	{

#ifdef ARCH_PC_LINUX
/*	### Bug in Linux 2.0.0 - inet STREAM sockets 

        if( domain == AF_INET && i > 30 ) break; */

#endif /* ARCH_PC_LINUX */

	    on = 1024*i;
	    onlen = sizeof(on);

 	    ret = setsockopt( Sessions[MAX_SESSIONS].mbox, SOL_SOCKET, SO_SNDBUF, (void *)&on, onlen);
	    if (ret < 0 ) break;

	    ret = setsockopt( Sessions[MAX_SESSIONS].mbox, SOL_SOCKET, SO_RCVBUF, (void *)&on, onlen);
	    if (ret < 0 ) break;

	    ret= getsockopt( Sessions[MAX_SESSIONS].mbox, SOL_SOCKET, SO_SNDBUF, (void *)&on, &onlen );
	    if( on < i*1024 ) break;
	    Alarm( SESSION, "Sess_accept: set sndbuf %d, ret is %d\n", on, ret );

	    onlen = sizeof(on);
	    ret= getsockopt( Sessions[MAX_SESSIONS].mbox, SOL_SOCKET, SO_RCVBUF, (void *)&on, &onlen );
	    if( on < i*1024 ) break;
	    Alarm( SESSION, "Sess_accept: set rcvbuf %d, ret is %d\n", on, ret );

	}
	Alarm( SESSION, "Sess_accept: set sndbuf/rcvbuf to %d\n", 1024*(i-5) );

        if ( domain == AF_INET ) {
                on = 1;
                ret = setsockopt( Sessions[MAX_SESSIONS].mbox, IPPROTO_TCP, TCP_NODELAY, (void *)&on, 4);
                if (ret < 0) 
                        Alarm(PRINT, "Setting TCP_NODELAY failed with  error: %s\n", strerror(errno));
                else
                        Alarm( SESSION, "Setting TCP_NODELAY on socket %d\n", Sessions[MAX_SESSIONS].mbox );
        }
	/* delaying for the private name to be written */
	Sess_detach_accept();
	E_attach_fd( Sessions[MAX_SESSIONS].mbox, READ_FD, Sess_accept_continue, 0, NULL, LOW_PRIORITY );	
	E_attach_fd( Sessions[MAX_SESSIONS].mbox, EXCEPT_FD, Sess_accept_continue, 0, NULL, LOW_PRIORITY );	
	accept_delay.sec = 1;
	accept_delay.usec= 0;
	E_queue( Sess_accept_continue2, 1, NULL, accept_delay );
}

void	Sess_accept_continue(mailbox d1, int d2, void *d3)
{
	char			response;
	int			legal_private_name;
	int			unique_private_name;
	char			private_group_name[MAX_GROUP_NAME];
	int			name_len;
	int			ioctl_cmd;
	char			ip[16];
        char                    version[3];
	char			conn[MAX_PRIVATE_NAME+2];
	int			ret, i, sess_location;

	E_dequeue( Sess_accept_continue2, 1, NULL );
	E_detach_fd( Sessions[MAX_SESSIONS].mbox, READ_FD );
	E_detach_fd( Sessions[MAX_SESSIONS].mbox, EXCEPT_FD );
	Sess_attach_accept();

	/* set file descriptor to non blocking */
	ioctl_cmd = 1;
	ret = ioctl( Sessions[MAX_SESSIONS].mbox, FIONBIO, &ioctl_cmd);

	/* 
	 * connect message looks like:
	 *
	 * byte - version of lib
	 * byte - subversion of lib
         * (optional) byte - patchversion of lib (only if version.subversion > 3.14)
	 * byte - 1/0 with or without groups
	 * byte - len of name
	 * len bytes - name
	 *
	 */
	/* version checking  3.01 is minimal */

        version[0] = version[1] = version[2] = 0;

	ret = recv( Sessions[MAX_SESSIONS].mbox, version, 2, 0 );
	if( ret < 0 )
	{
		Alarm( SESSION, "Sess_accept_continue: reading version.subversion failed on mailbox %d\n", 
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
        }
        if ( version[0]*10000 + version[1]*100 + version[2] > 31400 )
        {
            ret = recv( Sessions[MAX_SESSIONS].mbox, &version[2], 1, 0 );
            if( ret < 0 )
            {
		Alarm( SESSION, "Sess_accept_continue: reading patch_version failed on mailbox %d\n", 
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
            }
        }
	if( version[0]*10000 + version[1]*100 + version[2] < 30100 ) 
	{
		response = REJECT_VERSION;
		send( Sessions[MAX_SESSIONS].mbox, &response, sizeof(response), 0 );
		Alarm( SESSION, "Sess_accept_continue: version %d.%d.%d is not supported\n",
			version[0], version[1], version[2] );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
	}

	/* reading private name */
	for( i=0; i < MAX_PRIVATE_NAME+1; i++ )
		Sessions[MAX_SESSIONS].name[i] = 0;

	ret = recv( Sessions[MAX_SESSIONS].mbox, conn, MAX_PRIVATE_NAME + 2, 0 );
	if( ret < 0 )
	{
		Alarm( SESSION, "Sess_accept_continue: reading private name failed on mailbox %d\n", 
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;

	}else if( ret < 3 ){
		response = REJECT_NO_NAME;
		send( Sessions[MAX_SESSIONS].mbox, &response, sizeof(response), 0 );
		Alarm( SESSION, "Sess_accept_continue: reading private name failed on mailbox %d\n",
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
	}

        if( ((int)conn[0] % 2) ==  1 ) Sessions[MAX_SESSIONS].status = Set_memb_session( Sessions[MAX_SESSIONS].status );
        else Sessions[MAX_SESSIONS].status = Clear_memb_session( Sessions[MAX_SESSIONS].status );
        Sessions[MAX_SESSIONS].priority = (int)conn[0] / 16 ;
          
	name_len = (int)conn[1];
	if( name_len+2 != ret || name_len > MAX_PRIVATE_NAME || name_len < 1 )
	{
		response = REJECT_ILLEGAL_NAME;
		send( Sessions[MAX_SESSIONS].mbox, &response, sizeof(response), 0 );
		Alarm( SESSION, "Sess_accept_continue: len %d of private name does not fit (ret = %d) on mailbox %d\n",
			name_len, ret,
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
	}
	memcpy( Sessions[MAX_SESSIONS].name, &conn[2], name_len );

	/* set file descriptor back to blocking */
	ioctl_cmd = 0;
	ret = ioctl( Sessions[MAX_SESSIONS].mbox, FIONBIO, &ioctl_cmd);

	/* checking if private name is legal */
	for( legal_private_name=1, i=0; i < name_len; i++ )
		if( Sessions[MAX_SESSIONS].name[i] <= '#' ||
		    Sessions[MAX_SESSIONS].name[i] >  '~' ) legal_private_name = 0;
	if( !legal_private_name )
	{
		response = REJECT_ILLEGAL_NAME;
		send( Sessions[MAX_SESSIONS].mbox, &response, sizeof(response), 0 );
		Alarm( SESSION, "Sess_accept_continue: illegal private name %s on mailbox %d\n",
			Sessions[MAX_SESSIONS].name,
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
	}

	/* checking if private name is unique */
	for( unique_private_name=1, sess_location=0; sess_location < Num_sessions; sess_location++ )
	{
		ret = strcmp( Sessions[MAX_SESSIONS].name, Sessions[sess_location].name );
		if( ret <= 0 )
		{
			if( ret == 0 ) unique_private_name = 0;
			break;
		}
	}
	if( !unique_private_name )
	{
		response = REJECT_NOT_UNIQUE;
		send( Sessions[MAX_SESSIONS].mbox, &response, sizeof(response), 0 );
		Alarm( SESSION, "Sess_accept_continue: non unique private name %s on mailbox %d\n",
			Sessions[MAX_SESSIONS].name,
			Sessions[MAX_SESSIONS].mbox );
		close( Sessions[MAX_SESSIONS].mbox );
		return;
	}

	/* 
	 * accept message looks like:
	 *
	 * byte - ACCEPT_SESSION code
	 * byte - version of spread
	 * byte - subversion of spread
         * (optional) byte - patch version of spread (only if library is 3.15.0 or greater)
	 * byte - len of name
	 * len bytes - name
	 *
	 */
	Sessions[MAX_SESSIONS].num_mess = 0;
	response = ACCEPT_SESSION;
	send( Sessions[MAX_SESSIONS].mbox, &response, 1, 0 );

	response = SP_MAJOR_VERSION;
	send( Sessions[MAX_SESSIONS].mbox, &response, 1, 0 );

	response = SP_MINOR_VERSION;
	send( Sessions[MAX_SESSIONS].mbox, &response, 1, 0 );

        if (version[0]*10000 + version[1]*100 + version[2] >= 31500)
        {
            response = SP_PATCH_VERSION;
            send( Sessions[MAX_SESSIONS].mbox, &response, 1, 0 );
        }

	sprintf(private_group_name, "#%s#%s", Sessions[MAX_SESSIONS].name, My.name );
	name_len = strlen( private_group_name );
	/* sending the len of the private group in one byte */
	response = name_len;
	send( Sessions[MAX_SESSIONS].mbox, &response, 1, 0 );
	/* sending the private group name */
	send( Sessions[MAX_SESSIONS].mbox, private_group_name, name_len, 0 );
	E_attach_fd( Sessions[MAX_SESSIONS].mbox, READ_FD, Sess_read, Sessions[MAX_SESSIONS].type, NULL, 
			LOW_PRIORITY );
	E_attach_fd( Sessions[MAX_SESSIONS].mbox, EXCEPT_FD, Sess_read, Sessions[MAX_SESSIONS].type, NULL, 
			LOW_PRIORITY );
	Sessions[MAX_SESSIONS].status = Set_op_session( Sessions[MAX_SESSIONS].status );

        Prot_Create_Local_Session(&Sessions[MAX_SESSIONS]);

        Message_reset_current_location(&(Sessions[MAX_SESSIONS].read) );
        Message_reset_current_location(&(Sessions[MAX_SESSIONS].write) );
        Sessions[MAX_SESSIONS].read.in_mess_head = 1;

	Log_sess_connect( Sessions[MAX_SESSIONS].mbox, Sessions[MAX_SESSIONS].address, 
			  Sessions[MAX_SESSIONS].name );

	Conf_id_to_str( Sessions[MAX_SESSIONS].address, ip );
	Alarm( SESSION, "Sess_accept_continue: Accepting from %s with private name %s on mailbox %d\n",
		ip,
		Sessions[MAX_SESSIONS].name,
		Sessions[MAX_SESSIONS].mbox );

	/* sorting the Sessions structure according to name */
	memmove( &Sessions[sess_location+1], &Sessions[sess_location], (Num_sessions - sess_location)*sizeof(session) );
	memmove( &Sessions[sess_location], &Sessions[MAX_SESSIONS], sizeof(session) );

	Num_sessions++;
	GlobalStatus.num_sessions = Num_sessions;

	/* updating Session_index structure */
	for( i=0; i < FD_SETSIZE; i++ ) Session_index[i] = -1;
	for( i=0; i < Num_sessions; i++ )
		Session_index[ Sessions[i].mbox ] = i;

}

static	void	Sess_read( mailbox mbox, int dummy, void *d2 )
{
	message_header	*head_ptr, *msg_head;
        message_obj     *msg;
        scatter         *scat;
	down_link	*down_ptr;
        message_link    *mess_link;
	int		packet_index, byte_index, to_read;
#ifdef  CHECK_SENDERS
	char		private_name[MAX_PRIVATE_NAME+1];
	char		proc_name[MAX_PROC_NAME];
#endif
	int             len, remain, ret;
        int             head_size, data_frag_len;
        int             ses, ioctl_cmd, i, legal_private_name;
        char            tmp_name[MAX_GROUP_NAME];
        char            *c_ptr;
#if 0
#ifndef ARCH_SCATTER_NONE
static  struct  msghdr  msgh;
#endif  /* ARCH_SCATTER_NONE */
#endif  /* 0 */
#if 0
        /* we currently don't use recvmsg */
#ifndef ARCH_SCATTER_NONE
	msgh.msg_name    = (caddr_t) 0;
	msgh.msg_namelen = 0;
	msgh.msg_iov     = (struct iovec *)scat->elements;
	msgh.msg_iovlen  = scat->num_elements;
#endif  /* ARCH_SCATTER_NONE */

#ifdef ARCH_SCATTER_CONTROL
	msgh.msg_control = (caddr_t) 0;
	msgh.msg_controllen = 0;
#endif /* ARCH_SCATTER_CONTROL */
#ifdef ARCH_SCATTER_ACCRIGHTS
	msgh.msg_accrights = (caddr_t) 0;
	msgh.msg_accrightslen = 0;
#endif /* ARCH_SCATTER_ACCRIGHTS */
#endif /* 0 */

        ses = Session_index[mbox];
        if (Sessions[ses].read_mess == NULL)
                Sessions[ses].read_mess = Message_new_message();

        msg = Sessions[ses].read_mess;
	head_ptr = Message_get_message_header(msg);
        head_size = Message_get_header_size();

        /* set file descriptor to non blocking */
	ioctl_cmd = 1;
	ret = ioctl( mbox, FIONBIO, &ioctl_cmd);

        if ( Sessions[ses].read.in_mess_head == 1 )
        {
                /* read up to size of message_header */
                len = Sessions[ses].read.cur_byte;
                remain = sizeof(message_header) - len;
                ret = recv( mbox, (char *) &head_ptr[len], remain, 0 );
                if( ret  == remain )
                {
                        Sessions[ses].read.cur_byte += ret;
                        Message_set_location_begin_body(&(Sessions[ses].read) );
                } else  if (ret > 0 ) {
                        Sessions[ses].read.cur_byte += ret;
                        ioctl_cmd = 0;
                        ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);        
                        return;
                } else {
                        /* error reading */
                        Alarm( SESSION, "Sess_read: failed receiving header on session %d: ret %d: error: %s \n", mbox, ret, strerror(errno) );
                        Sess_kill( mbox );
                        ioctl_cmd = 0;
                        ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);        
                        return;
                }
                /* When we get here we have a complete header */
                ioctl_cmd = 0;
                ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);        

                /* Fliping message header to my form if needed */
                if( !Same_endian( head_ptr->type ) ) 
                {
                        Flip_mess( head_ptr );
                }
                /* Setting endian to my endian on the header */
                head_ptr->type = Set_endian( head_ptr->type );

                /* Validate all fields */
                Alarm( SESSION, "Sess_read: Message has type field 0x%x\n", head_ptr->type);
                if ( head_ptr->type & ~( UNRELIABLE_MESS | RELIABLE_MESS | FIFO_MESS | CAUSAL_MESS | 
                                         AGREED_MESS | SAFE_MESS | SELF_DISCARD | 
                                         DROP_RECV | JOIN_MESS | LEAVE_MESS | KILL_MESS | 0x80000080 ) )
                {
                        Alarm( SESSION, "Sess_read: Message has illegal type field 0x%x\n", head_ptr->type);
                        Sess_kill(mbox);
                        return;
                }
                head_ptr->private_group_name[MAX_GROUP_NAME -1] = '\0';
                memcpy(tmp_name, head_ptr->private_group_name, MAX_GROUP_NAME);
                c_ptr = strchr(&tmp_name[1], '#');
                if (c_ptr != NULL)
                        c_ptr[0] = '\0';
                if ( (c_ptr == NULL) || (tmp_name[0] != '#' ) || (strlen(&(tmp_name[1])) > MAX_PRIVATE_NAME) ||
                     ( strlen(&(c_ptr[1])) >= MAX_PROC_NAME ) )
                {
                        Alarm( SESSION, "Sess_read: Message has illegal private_group_name (priv, proc)\n");
                        Sess_kill(mbox);
                        return;
                }
                legal_private_name = 1;
                for( i=1; i < strlen(&tmp_name[1]); i++ )
                        if( tmp_name[i] <= '#' ||
                            tmp_name[i] >  '~' ) legal_private_name = 0;
                for( i=1; i < strlen(&c_ptr[1]); i++ )
                        if( c_ptr[i] <= '#' ||
                            c_ptr[i] >  '~' ) legal_private_name = 0;
                if( !legal_private_name )
                {
                        Alarm( SESSION, "Sess_read: Message has illegal private_group_name characters %s\n", tmp_name );
                        Sess_kill(mbox);
                        return;
                }
                if ( (head_ptr->num_groups < 0) || (head_ptr->num_groups > MAX_GROUPS_PER_MESSAGE) )
                {
                        Alarm( SESSION, "Sess_read: Message has negative or too large num_groups field\n", head_ptr->num_groups);
                        Sess_kill(mbox);
                        return;
                }
                if ( head_ptr->hint & ~( 0x80000080 | 0x00ffff00 ) )
                {
                        Alarm( SESSION, "Sess_read: Message has illegal hint field 0x%x\n", head_ptr->hint);
                        Sess_kill(mbox);
                        return;
                }

                if ( (head_ptr->data_len < 0 ) || ( head_ptr->data_len > (MAX_SCATTER_ELEMENTS * sizeof(packet_body))) )
                {
                        Alarm( SESSION, "Sess_read: Message has negative or too large data_len %d\n", head_ptr->data_len);
                        Sess_kill(mbox);
                        return;
                }
                if ( (head_ptr->data_len + MAX_GROUP_NAME * head_ptr->num_groups) > (MAX_SCATTER_ELEMENTS * Message_get_data_fragment_len() - head_size ) )
                {
                        Alarm( SESSION, "Sess_read: Message + Groups is too large (%d + %d = %d). MAX size is: %d\n", 
                               head_ptr->data_len, MAX_GROUP_NAME * head_ptr->num_groups, 
                               head_ptr->data_len + MAX_GROUP_NAME * head_ptr->num_groups,
                               MAX_SCATTER_ELEMENTS * Message_get_data_fragment_len() - head_size );
                        Sess_kill(mbox);
                        return;
                }
        } /* finished reading and validating  header */

	/* 
	 * to do recvmsg, but need to trick with the starting AFTER the header 
	 * on the first packet, and then to return the big_scatter to original
	 * form. read at *most* head_ptr->data_len (set scatter lengths accordingly
	 * everytime here from scratch! 
	 *
	 * ret = recvmsg( mbox, &msg, 0 ); 
	 * if( ret <=0 )
	 * {
	 * 	Alarm( SESSION, "Sess_read: failed receiving message on session %d\n", mbox );
	 * 	Sess_kill( mbox );
	 * 	return;
	 * }
	 */

	/* read the rest of the message if needed, reserving room at the beginning
         * of the first fragment(scat buf) for the message header and the lts and seq fields. */

        /* enable non-blocking io */
	ioctl_cmd = 1;
	ret = ioctl( mbox, FIONBIO, &ioctl_cmd);

        data_frag_len = Message_get_data_fragment_len();
        scat = Message_get_data_scatter(msg);
	remain = ( head_ptr->data_len + MAX_GROUP_NAME*head_ptr->num_groups )  - Sessions[ses].read.total_bytes;
	for(  ; remain > 0; remain -= ret )
	{
		packet_index = Sessions[ses].read.cur_element;
		byte_index   = Sessions[ses].read.cur_byte;
                if (packet_index >= scat->num_elements) 
                {
                        /* We are beginning a new fragment -- so allocate it */
                        assert(byte_index == 0);
                        Message_add_scat_element(msg);
                }
		to_read = ( data_frag_len - byte_index );
		if( to_read > remain ) to_read = remain;
		ret = recv( mbox, &scat->elements[packet_index].buf[byte_index],
				to_read, 0 );
                if( ret  == to_read )
                {
                        Sessions[ses].read.cur_byte = 0;
                        Sessions[ses].read.cur_element++;
                        Sessions[ses].read.total_bytes += ret;
                } else  if (ret > 0 ) {
                        Sessions[ses].read.cur_byte += ret;
                        Sessions[ses].read.total_bytes += ret;
                        ioctl_cmd = 0;
                        ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);        
                        return;
                } else {
			Alarm( SESSION, "Sess_read: failed receiving message on session %d, ret is %d: error: %s\n", mbox, ret, strerror(errno) );
			Alarm( SESSION, "Sess_read: failed recv msg more info: len read: %d, remain: %d, to_read: %d, pkt_index: %d, b_index: %d, scat_nums: %d\n",Sessions[ses].read.total_bytes, remain, to_read, packet_index, byte_index, scat->num_elements );
			Sess_kill( mbox );
                        ioctl_cmd = 0;
                        ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);        
			return;
		}
	}

        /* We now have a complete message */
        ioctl_cmd = 0;
        ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);        

        /* reset active read_mess to empty */
        Message_reset_current_location(&(Sessions[ses].read));
        Sessions[ses].read.in_mess_head = 1;
        Sessions[ses].read_mess = NULL;


#ifdef  PROBE_LATENCY
        if (Is_latency_mess( head_ptr->type ) )
        {
                int32u          initial_offset, htime_offset;
                int32u          *p_time_offset;
                sp_time         cur_time, ncur_time;

                initial_offset = MAX_GROUP_NAME*head_ptr->num_groups + head_size;
                p_time_offset = (int32u *) &msg->body.elements[0].buf[initial_offset];
                htime_offset = ntohl(*p_time_offset);
                Alarm(SESSION, "Sess_read: Msg Data at %d with time_offset %u \n", initial_offset, htime_offset);
                cur_time = E_get_time();
                ncur_time.sec = htonl(cur_time.sec);
                ncur_time.usec = htonl(cur_time.usec);
                memcpy(&(msg->body.elements[0].buf[htime_offset + initial_offset]), &ncur_time, sizeof(sp_time));
                Alarm(SESSION, "Sess_read: timestamped time (%d, %d) in byte %d of message\n", cur_time.sec, cur_time.usec, htime_offset);
                *p_time_offset = htonl(htime_offset + sizeof(sp_time) );
        }
#endif  /* PROBE_LATENCY */
#ifdef  CHECK_SENDERS
	/* 
	 * ### if you want to check validity of sender
	 * ret = G_private_to_names( head_ptr->private_group_name, private_name, proc_name );
	 * if( ret < 0 )
	 * {
	 * 	Alarm( PRINT, "Sess_read: Illegal private name %s\n",
	 * 			head_ptr->private_group_name );
	 * 	return;
	 * }
	 * if( strcmp( proc_name, My.name ) != 0 )
	 * {
	 * 	Alarm( PRINT, "Sess_read: proc name %s is not my name %s\n",
	 * 			proc_name, My.name );
	 * 	return;
	 * }
	 * if( strcmp( private_name, Sessions[ Session_index[ mbox ] ].name ) != 0 )
	 * {
	 * 	Alarm( PRINT, "Sess_read: session name %s and private name %s do not match\n",
	 * 			head_ptr->private_group_name, 
	 * 	    		Sessions[ Session_index[ mbox ] ].name );
	 * 	return;
	 * }
	 */
#endif


        if ( Is_leave_mess(head_ptr->type) || Is_join_mess(head_ptr->type) || Is_kill_mess(head_ptr->type) )
        {
                memcpy(tmp_name, head_ptr->private_group_name, MAX_GROUP_NAME);
                c_ptr = strchr(&tmp_name[1], '#');
                if (c_ptr != NULL)
                        c_ptr[0] = '\0';
                if (strncmp(&tmp_name[1], Sessions[ses].name, MAX_PRIVATE_NAME) )
                {
                        Alarm( EXIT, "Sess_read: Session %s trying to make session %s do something\n",
                               &tmp_name[1], Sessions[ses].name );
			Sess_kill( mbox );
                        return;
                }
        }

	/* create new down_link and big_scatter */
        down_ptr = Prot_Create_Down_Link(msg, Message_get_packet_type(head_ptr->type), mbox, 0);
        if (down_ptr == NULL)
        {
                Alarm( SESSION, "Sess_read: Session has illegal message type 0x%x\n", head_ptr->type);
                Sess_kill( mbox );
                return;
        }
        down_ptr->mess = msg;
        Message_element_len_fixup(down_ptr->mess);
        msg_head = Message_get_message_header(down_ptr->mess);

        if (Is_kill_mess(msg_head->type) )
        {       
                /* We are going to overwrite the group that is sent from the library. 
                 * it is not needed for kill messages, so we shrink the data field to ignore it
                 */
                len = Message_kill_mess_fixup(down_ptr->mess, Sessions[ses].read.total_bytes - MAX_GROUP_NAME, mbox);

                /* A bug in both 3.13 and 4 I think is that if we get a DISCONNECT message
                 * from the client and we process and deliver that before discovering the
                 * closed socket ourselves and calling Sess_kill(), then the session
                 * is in the wrong state and we will crash when we try to finish delivery.
                 */
                Log_sess_disconnect( Sessions[ses].mbox, Sessions[ses].address, Sessions[ses].name,
                                     Sessions[ses].num_mess );
                /* clear his structure */
                while( Sessions[ses].num_mess > 0 )
                {
                        mess_link = Sessions[ses].first;
                        Sessions[ses].first = Sessions[ses].first->next;
                        Sess_dispose_message( mess_link );
                        Sessions[ses].num_mess--;
                }

                /* close the mailbox and mark it unoperational */
                E_dequeue( Sess_badger, mbox, NULL );
                E_detach_fd( mbox, READ_FD );
                E_detach_fd( mbox, EXCEPT_FD );
                close( mbox );
                /* the mailbox is closed but the entry still points to it */
                Sessions[ses].status = Clear_op_session( Sessions[ses].status );
                Alarm( SESSION, "Sess_read: disconnecting session %s ( mailbox %d )\n",Sessions[ses].name, mbox );
        }

	Alarm( SESSION, "Sess_read: queueing message of type %d with len %d to the protocol\n",
		down_ptr->type, Sessions[ses].read.total_bytes );
	Prot_new_message( down_ptr, Sessions[ses].down_queue );
}

void    Sess_write( int ses, message_link *mess_link, int *needed )
{
        message_obj     *msg;
	message_link	*tmp_link;
        scatter         *scat;
	int		ioctl_cmd;
	int 		ret;
	int		total_to_send;
	int		len_sent, first_data_len;
        char            *first_data_ptr;
	int		i;
        message_header  *head_ptr;

	if( !Is_op_session( Sessions[ses].status ) ) return;

	if( Sessions[ses].num_mess >= MAX_SESSION_MESSAGES )
	{
		Alarm( SESSION, 
			"Sess_write: killing mbox %d for not reading\n",
			Sessions[ses].mbox );
		Sess_kill( Sessions[ses].mbox );
		return;
	}

	if( Sessions[ses].num_mess > 0 )
	{
		Sess_badger( Sessions[ses].mbox , NULL);
	}

	msg = mess_link->mess;
        Obj_Inc_Refcount(msg);
        scat = Message_get_data_scatter(msg);

	for( total_to_send=0, i=0; i < scat->num_elements; i++ )
		total_to_send += scat->elements[i].len;

        /* since also sending message_header */
        total_to_send += Message_get_non_body_header_size();

        head_ptr = Message_get_message_header(msg);
#ifdef  PROBE_LATENCY
        if (Is_latency_mess( head_ptr->type ) )
        {
                int32u          initial_offset, htime_offset;
                int32u          *p_time_offset;
                sp_time         cur_time, ncur_time;

                initial_offset = MAX_GROUP_NAME*head_ptr->num_groups;
                p_time_offset = (int32u *) &msg->body.elements[0].buf[initial_offset];
                htime_offset = ntohl(*p_time_offset);
                Alarm(SESSION, "Sess_write: Msg Data at %d with timeoffset %u\n", initial_offset, htime_offset);
                cur_time = E_get_time();
                ncur_time.sec = htonl(cur_time.sec);
                ncur_time.usec = htonl(cur_time.usec);
                memcpy(&(msg->body.elements[0].buf[htime_offset + initial_offset]), &ncur_time, sizeof(sp_time));
                Alarm(SESSION, "Sess_write: timestamped time (%d, %d) in byte %d of message\n", cur_time.sec, cur_time.usec, htime_offset);
                *p_time_offset = htonl(htime_offset + sizeof(sp_time) );
        }
#endif
	len_sent = 0;
	if( Sessions[ses].num_mess == 0 )
	{
		/* set file descriptor to non blocking */
		ioctl_cmd = 1;
		ret = ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);

                /* Send the first part of message which must be handled specially */
                ret = send(Sessions[ses].mbox, (char *) head_ptr, 
                           sizeof(message_header), 0);
                if ( ret > 0 ) len_sent += ret;
                if ( ret != sizeof(message_header) )
                {
                        goto end_write;
                }
                /* Send the first data of message which must be handled specially */
                first_data_ptr = Message_get_first_data_ptr(msg);
                first_data_len = Message_get_first_data_len(msg);
                ret = send(Sessions[ses].mbox, first_data_ptr, 
                           first_data_len, 0);
                if ( ret > 0 ) len_sent += ret;
                if ( ret != first_data_len )
                {
                        goto end_write;
                }
		/* send the message after first buffer*/
		for( i=1; i < scat->num_elements; i++ )
		{
			ret = send( Sessions[ses].mbox, scat->elements[i].buf, 
							scat->elements[i].len, 0);
			if( ret > 0 ) len_sent += ret;
			if( ret != scat->elements[i].len )
			{ 
				break;
			}
		}
		/* set file descriptor back to blocking */
        end_write:
		ioctl_cmd = 0;
		ret = ioctl( Sessions[ses].mbox, FIONBIO, &ioctl_cmd);

	}

	if( len_sent < total_to_send )
	{
		/* this message has to be linked */
		if( *needed )
		{
			/* create a copy of mess_link and link it */
			tmp_link = new(MESSAGE_LINK);
                        if (tmp_link == NULL ) {
                                Alarm(EXIT, "Sess_write: Failed to allocate a new MESSAGE_LINK.\n");
                                return;
                        }
                        tmp_link->mess = Message_copy_message(msg);
			++*needed;
		}else{
			/* should link mess_link itself */
			tmp_link = mess_link;
			*needed=1;
		}
		/* link the message */
		tmp_link->next = 0;
		if( Sessions[ses].num_mess == 0 )
		{
			Sessions[ses].first = tmp_link;
			Sessions[ses].last = tmp_link;
			/* setting cur_element and cur_byte */
                        Message_calculate_current_location(tmp_link->mess, len_sent, &(Sessions[ses].write) );

			/* We will need to badger this guy */
			E_queue( Sess_badger, Sessions[ses].mbox, NULL, Badger_timeout );
		}else{
			/* This guy was already badgered */
			Sessions[ses].last->next = tmp_link;
			Sessions[ses].last = tmp_link;
		}
		Sessions[ses].num_mess++;
	}
        Message_Dec_Refcount(msg);
}

static	void	Sess_badger( mailbox mbox, void *dummy )
{
	int		ses;
	message_link	*mess_link;
	int		able_to_write;
        message_obj     *msg;
        scatter         *scat;
	int		ioctl_cmd;
	int		bytes_to_send, from;
	int		i;
	int		ret;

	Alarm( SESSION, "Sess_badger: for mbox %d\n", mbox );
	ses = Session_index[ mbox ];
	if( ses < 0 || ses >= Num_sessions ) return;
	if( !Is_op_session( Sessions[ses].status ) ) return;

	if( Sessions[ses].num_mess <= 0 ) return;

	/* set file descriptor to non blocking */
	ioctl_cmd = 1;
	ret = ioctl( mbox, FIONBIO, &ioctl_cmd);

	for( able_to_write = 1 ; Sessions[ses].num_mess > 0 && able_to_write;  )
	{
		msg = Sessions[ses].first->mess;
                Obj_Inc_Refcount(msg);

                /* First if we havn't sent the mess_head yet, send that special
                 * and then reset the cur_byte and in_mess_head fields 
                 */
                if ( Sessions[ses].write.in_mess_head == 1 )
                {
                        char *tmp_buf;

                        tmp_buf = (char *) Message_get_message_header(msg);
                        from = Sessions[ses].write.cur_byte;
                        bytes_to_send = sizeof(message_header) - from;
                        ret = send(mbox, &tmp_buf[from], bytes_to_send, 0);
                        if (ret == bytes_to_send)
                        {
                                Message_set_location_begin_body(&(Sessions[ses].write));
                        } else 
                        {
                                if (ret > 0) Sessions[ses].write.cur_byte = from + ret;
                                able_to_write = 0;
                        }
                }
                /* Next if we still can, send some body of this messsage */
                if ( able_to_write )
                {
                        scat = Message_get_data_scatter(msg);
                        for( i=Sessions[ses].write.cur_element, from=Sessions[ses].write.cur_byte;
                             i < scat->num_elements; 
                             i++, from=0 )
                        {
                                bytes_to_send = scat->elements[i].len - from;
                                ret = send( mbox, &(scat->elements[i].buf[from]), bytes_to_send, 0);
                                if( ret == bytes_to_send )
                                {        
                                        Sessions[ses].write.cur_byte = 0;
                                        Sessions[ses].write.cur_element++;
                                }else{
                                        if( ret > 0 ) Sessions[ses].write.cur_byte += ret;
                                        able_to_write = 0;
                                        break; 
                                }
                        }
                }
		if( able_to_write )
		{
			/* free that message */
			mess_link = Sessions[ses].first;
			Sessions[ses].first = Sessions[ses].first->next;
			Sessions[ses].num_mess--;
                        Message_reset_current_location(&(Sessions[ses].write) );
			Sess_dispose_message( mess_link );
		}
                Message_Dec_Refcount(msg);
	} /* for loop per message */
	/* set file descriptor back to blocking */
	ioctl_cmd = 0;
	ret = ioctl( mbox, FIONBIO, &ioctl_cmd);

	if( Sessions[ses].num_mess > 0 ) E_queue( Sess_badger, mbox, NULL, Badger_timeout );
}

static	void	Sess_kill( mailbox mbox )
{
	int		ses;
	message_link	*mess_link;
	message_obj	*kill_mess;
        message_header  *kill_head;
	down_link	*down_ptr;
        int             head_size;
        char            private_send_group[MAX_GROUP_NAME];

        head_size = Message_get_header_size();
	ses = Session_index[mbox];
	if( !Is_op_session( Sessions[ses].status ) )
		Alarm( PRINT, 
		  "Sess_kill: session %s with mailbox %d is already killed (status %d)\n",
		  Sessions[ses].name, mbox, Sessions[ses].status );

	/* send a message to kill this session */
	sprintf( private_send_group, "#%s#%s", Sessions[ses].name, My.name );
        kill_mess = Message_create_message(KILL_MESS | SAFE_MESS, private_send_group);
        kill_head = Message_get_message_header(kill_mess);
        Message_kill_mess_fixup(kill_mess, head_size, mbox );

        down_ptr = Prot_Create_Down_Link(kill_mess, Message_get_packet_type(kill_head->type), mbox, 0);
	down_ptr->mess = kill_mess;

        Obj_Inc_Refcount(down_ptr->mess);
	Prot_new_message( down_ptr, Sessions[ses].down_queue );
        Message_Dec_Refcount(kill_mess);

	Log_sess_disconnect( Sessions[ses].mbox, Sessions[ses].address, Sessions[ses].name,
			     Sessions[ses].num_mess );
	/* clear his structure */
	while( Sessions[ses].num_mess > 0 )
	{
		mess_link = Sessions[ses].first;
		Sessions[ses].first = Sessions[ses].first->next;
		Sess_dispose_message( mess_link );
		Sessions[ses].num_mess--;
	}
        /* reset active read_mess to empty */
        Message_reset_current_location(&(Sessions[ses].read));
        Sessions[ses].read.in_mess_head = 1;
        Sessions[ses].read_mess = NULL;

	/* close the mailbox and mark it unoperational */
	E_dequeue( Sess_badger, mbox, NULL );
	E_detach_fd( mbox, READ_FD );
	E_detach_fd( mbox, EXCEPT_FD );
	close( mbox );
	/* the mailbox is closed but the entry still points to it */
	Sessions[ses].status = Clear_op_session( Sessions[ses].status );
	Alarm( SESSION, "Sess_kill: killing session %s ( mailbox %d )\n",Sessions[ses].name, mbox );
}

void	Sess_deliver_message( message_link *mess_link )
{
static	int		target_sessions[MAX_SESSIONS];
	int		num_target_sessions;
	int		source_ses;
	message_header	*head_ptr;
        message_obj     *msg;
        scatter         *scat;
	char		*target_groups;
	char		private_name[MAX_PRIVATE_NAME+1];
	char		proc_name[MAX_PROC_NAME];
	int		groups_bytes, num_bytes;
	int		needed;
	int		i;

	msg = mess_link->mess;
        Obj_Inc_Refcount(msg);
	head_ptr = Message_get_message_header(msg);

        Message_endian_correct_message_header(msg);

	if( Is_join_mess( head_ptr->type ) )
	{
		Sess_handle_join( mess_link );
                Message_Dec_Refcount(msg);
		return;
	}
	
	if( Is_leave_mess( head_ptr->type ) )
	{
		Sess_handle_leave( mess_link );
                Message_Dec_Refcount(msg);
		return;
	}

	if( Is_kill_mess( head_ptr->type ) )
	{
		Sess_handle_kill( mess_link );
                Message_Dec_Refcount(msg);
		return;
	}
	
	if( Is_groups_mess( head_ptr->type ) )
	{
		G_handle_groups( mess_link );
                Message_Dec_Refcount(msg);
		return;
	}

	/* regular message */

	GlobalStatus.message_delivered++;

	/* Setting endian to my endian on the header */
	head_ptr->type = Set_endian( head_ptr->type );
        scat = Message_get_data_scatter(msg);
	/* analyze message  groups to sessions  */
        groups_bytes = Message_get_data_header_size();
	groups_bytes += (head_ptr->num_groups*MAX_GROUP_NAME);
	if( groups_bytes > scat->elements[0].len )
	{
		num_bytes = 0;
		for( i=0; i < scat->num_elements && num_bytes < groups_bytes; i++ )
		{ 
                    int copy_bytes;
                    if (groups_bytes - num_bytes < scat->elements[i].len)
                        copy_bytes = groups_bytes - num_bytes;
                    else 
                        copy_bytes = scat->elements[i].len;
                    memcpy( &Temp_buf[num_bytes], scat->elements[i].buf,  
                            copy_bytes );
                    num_bytes += copy_bytes;
		}
		target_groups = &Temp_buf[0];
	}else 	target_groups = Message_get_groups_starting_pointer(msg);
	num_target_sessions = G_analize_groups( head_ptr->num_groups, 
		(char (*)[MAX_GROUP_NAME])target_groups, target_sessions ) ;

	/* if self_discard, sender is local and a target then eliminate sender from targets */
	source_ses = -1;
	if( num_target_sessions > 0 && Is_self_discard( head_ptr->type ) )
	{
		G_private_to_names( head_ptr->private_group_name, private_name, proc_name );
		if( strcmp( My.name, proc_name ) == 0 )
		{
			source_ses = Sess_get_session( private_name );
		}
	}
	needed = 0;
	for( i = 0; i < num_target_sessions ; i++ )
	{
		if( source_ses == target_sessions[i] ) continue; /* self_discard */
		Sess_write( target_sessions[i], mess_link, &needed );
	}

        Message_Dec_Refcount(msg);
	if( !needed ) Sess_dispose_message( mess_link );
}

void	Sess_deliver_reg_memb( configuration reg_memb, membership_id reg_memb_id )
{
	G_handle_reg_memb( reg_memb, reg_memb_id );
}

void	Sess_deliver_trans_memb( configuration trans_memb )
{
	G_handle_trans_memb( trans_memb );
}

static	void	Sess_handle_join( message_link *mess_link )
{
	message_header	*join_head;
	char		*group_name;

	join_head = Message_get_message_header(mess_link->mess);
	group_name = Message_get_groups_starting_pointer(mess_link->mess);

	G_handle_join( join_head->private_group_name, group_name );

	Sess_dispose_message( mess_link );

}

static	void	Sess_handle_leave( message_link *mess_link )
{
	message_header	*leave_head;
	char		*group_name;

	leave_head = Message_get_message_header(mess_link->mess);
	group_name = Message_get_groups_starting_pointer(mess_link->mess);

	G_handle_leave( leave_head->private_group_name, group_name );

	Sess_dispose_message( mess_link );
}

static	void	Sess_handle_kill( message_link *mess_link )
{
	char		private_name[MAX_PRIVATE_NAME+1];
	char		proc_name[MAX_PROC_NAME];
	message_header	*kill_head;
	int		ses;
	int		i, ret;

	kill_head = Message_get_message_header(mess_link->mess);

	ret = G_private_to_names( kill_head->private_group_name, private_name, proc_name );
	if( ret < 0 ) Alarm( EXIT, "Sess_handle_kill: Illegal private name to kill %s\n", 
				kill_head->private_group_name );

	G_handle_kill( kill_head->private_group_name );

	if( strcmp( My.name, proc_name ) == 0 )
	{
		/* my machine, we need to find the session */
		ses = Sess_get_session( private_name );
		if( ses >= 0 )
		{
			/* delete session ses */
			if( Is_op_session( Sessions[ses].status ) )
				Alarm( EXIT, "Sess_handle_kill: killing unkilled session bug!\n");
                        Prot_Destroy_Local_Session(&(Sessions[ses]) );
			for( i=ses+1; i < Num_sessions; i++ )
				memmove( &Sessions[i-1], &Sessions[i], sizeof(session) );
			Num_sessions--;
			GlobalStatus.num_sessions = Num_sessions;

			/* updating Session_index structure */
			for( i=0; i < FD_SETSIZE; i++ ) Session_index[i] = -1;
			for( i=0; i < Num_sessions; i++ )
				Session_index[ Sessions[i].mbox ] = i;
		}
	}
        Prot_kill_session(mess_link->mess);

	Sess_dispose_message( mess_link );
}

void	Sess_dispose_message( message_link *mess_link )
{
        Message_dispose_message(mess_link->mess);
	dispose( mess_link );
}

int	Sess_get_session( char *name )
{
	int	i;
	int	ret;

	for( i=0; i < Num_sessions; i++ )
	{
		ret = strcmp( Sessions[i].name, name );
		if( ret <  0 ) continue;
		if( ret == 0 ) return( i );
		if( ret >  0 ) return( -1 );
	}
	return( -1 );
}

void    Flip_mess( message_header *head_ptr )
{
	head_ptr->type		= Flip_int32( head_ptr->type );
	head_ptr->num_groups	= Flip_int32( head_ptr->num_groups );
	head_ptr->data_len	= Flip_int32( head_ptr->data_len );
}
