/*
 * The contents of this file are subject to the Spread Non-Commercial 
 * License, Version 1.3 (the ``License''); you may not use
 * this file except in compliance with the License.  You may obtain a
 * copy of the License at:
 *
 * http://www.spread.org/license/
 *
 * or in the file ``license.txt'' found in this distribution.
 *
 * Software distributed under the License is distributed on
 * an AS IS basis, WITHOUT WARRANTY OF ANY KIND,
 * either express or implied. See the License for the
 * specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is:
 *    The Spread Wide Area Group Communication Toolkit.
 *     
 * The Creators of the Original Code are:
 *    Yair Amir, Michal Miskin-Amir, Jonathan Stanton.
 *
 *    Portions created by the Initial Developers are
 *    Copyright (C) 1993-2000 Spread Concepts LLC <spread@spreadconcepts.com>
 *
 *    All Rights Reserved.
 *
 * Major Contributor(s):
 * ---------------
 *    Dan Schoenblum   dansch@cnds.jhu.edu - Java Interface Developer.
 *    John Schultz     jschultz@cnds.jhu.edu - contribution to process group membership.
 *    Theo Schlossnagle theos@cnds.jhu.edu - Perl library and Skiplists.
 *
 */


#include <string.h>
#include "arch.h"
#include "spread_params.h"
#include "session.h"
#include "configuration.h"
#include "sp_events.h"
#include "status.h"
#include "log.h"
#include "alarm.h"

#ifndef ARCH_PC_WIN95
#include <grp.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#endif

#ifdef	ARCH_PC_WIN95

#include	<winsock.h>

WSADATA		WSAData;

#endif	/* ARCH_PC_WIN95 */

static	char		*My_name;
static	char		My_name_buf[80];
static	char		Config_file[80];
static	int		Log;

static	void	Usage(int argc, char *argv[]);

int main(int argc, char *argv[])
{
#ifdef	ARCH_PC_WIN95
	int	ret;
#endif

#ifndef ARCH_PC_WIN95
	struct group  *grp;
	struct passwd *pwd;
#endif

	Alarm_set( CONF ); 

	Alarm( PRINT, "/===========================================================================\\\n");
	Alarm( PRINT, "| The Spread Group Communication Toolkit.                                   |\n");
	Alarm( PRINT, "| Copyright (c) 1993-2000 Spread Concepts LLC                               |\n"); 
	Alarm( PRINT, "| All rights reserved.                                                      |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| The Spread package is licensed under the Spread Non-Commercial License.   |\n");
	Alarm( PRINT, "| You may only use this software in compliance with the License.            |\n");
	Alarm( PRINT, "| A copy of the license can be found at http://www.spread.org/license       |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| This software is distributed on an \"AS IS\" basis, WITHOUT WARRANTY OF     |\n");
	Alarm( PRINT, "| ANY KIND, either express or implied.                                      |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| Spread is developed at the Center for Networking and Distributed Systems, |\n");
	Alarm( PRINT, "| The Johns Hopkins University.                                             |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| Creators:                                                                 |\n");
	Alarm( PRINT, "|    Yair Amir             yairamir@cs.jhu.edu                              |\n");
	Alarm( PRINT, "|    Michal Miskin-Amir    michal@spread.org                                |\n");
	Alarm( PRINT, "|    Jonathan Stanton      jonathan@cs.jhu.edu                              |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| Major Contributors:                                                       |\n");
	Alarm( PRINT, "|    Dan Schoenblum   dansch@cnds.jhu.edu - Java Interface Developer.       |\n");
        Alarm( PRINT, "|    John Schultz     jschultz@cnds.jhu.edu - contribution to process group |\n");
        Alarm( PRINT, "|                                             membership.                   |\n");	
        Alarm( PRINT, "|    Theo Schlossnagle theos@cnds.jhu.edu - Perl library and Skiplists      |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| Special thanks to the following for providing ideas and/or code:          |\n");
	Alarm( PRINT, "|    Ken Birman, Danny Dolev, Mike Goodrich, Ben Laurie,                    |\n");
        Alarm( PRINT, "|    David Shaw, Robbert VanRenesse.                                        |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| For a full list of contributors, see Readme.txt in the distribution.      |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| WWW    : http://www.spread.org  and  http://www.cnds.jhu.edu              |\n");
	Alarm( PRINT, "| Contact: spread@spread.org                                                |\n");
	Alarm( PRINT, "|                                                                           |\n");
	Alarm( PRINT, "| Version %d.%02d.%02d Built 20/Mar/2001                                         |\n", 
		(int)SP_MAJOR_VERSION, (int)SP_MINOR_VERSION, (int)SP_PATCH_VERSION );
	Alarm( PRINT, "\\===========================================================================/\n");

	Usage( argc, argv );

#ifdef	ARCH_PC_WIN95

	ret = WSAStartup( MAKEWORD(1,1), &WSAData );
	if( ret != 0 )
		Alarm( EXIT, "Spread: winsock initialization error %d\n", ret );

#endif	/* ARCH_PC_WIN95 */

	Conf_init( Config_file, My_name );

	E_init();

	Sess_init();

	Stat_init(); 
	if( Log ) Log_init();

#ifndef	ARCH_PC_WIN95

	/* Yupp, we're paranoid */
 
	if (geteuid()) {
                Alarm( PRINT, "Spread: not running as root, won't chroot\n" );
	}
	else if ( (grp = getgrnam(SP_GROUP)) == NULL
                  || (pwd = getpwnam(SP_USER)) == NULL
                  || chdir(SP_RUNTIME_DIR) < 0
                  || chroot(SP_RUNTIME_DIR) < 0
                  || setgroups(0, 0) < 0
                  || setgid(grp->gr_gid) < 0
                  || setuid(pwd->pw_uid) < 0) 
        {
                Alarm( PRINT, "Spread: FAILED -- chroot or privledge drop -- check spread_params.h for directory and user/group names\n");
                Alarm( EXIT, "Spread: FAILED to give up privileges or chroot\n");
	} else {
                Alarm( PRINT, "Spread: setugid and chroot successeful\n" );
	}

#endif	/* ARCH_PC_WIN95 */

	E_handle_events();

	return 0;
}

static	void	Usage(int argc, char *argv[])
{
	My_name = 0; /* NULL */
	Log	= 0;
	strcpy( Config_file, "spread.conf" );

	while( --argc > 0 )
	{
		argv++;

		if( !strncmp( *argv, "-n", 2 ) )
		{
			if( strlen( argv[1] ) > MAX_PROC_NAME-1 ) /* -1 for the null */
				Alarm( EXIT, "Usage: proc name %s too long\n",
					argv[1] );

			memcpy( My_name_buf, argv[1], strlen( argv[1] ) );
			My_name = My_name_buf;

			argc--; argv++;

		}else if( !strncmp( *argv, "-c", 2 ) ){
			strcpy( Config_file, argv[1] );

			argc--; argv++;

		}else if( !strncmp( *argv, "-l", 2 ) ){

			if( !strcmp( argv[1], "y" ) )
				Log = 1;
			else if( !strcmp( argv[1], "n" ) )
				Log = 0;
			else Alarm( EXIT, "Usage: spread\n%s\n%s\n%s\n",
				"\t[-l y/n]          : print log",
				"\t[-n <proc name>]  : force computer name",
				"\t[-c <file name>]  : specify configuration file" );

			argc--; argv++;

		}else{
			Alarm( EXIT, "Usage: spread\n%s\n%s\n%s\n",
				"\t[-l y/n]          : print log",
				"\t[-n <proc name>]  : force computer name",
				"\t[-c <file name>]  : specify configuration file" );
		}
	}
}
